from . import exceptions


class Receiver:
    def receive_context(self, context):
        pass

    def receive_cleaned_context(self, context):
        pass

    def receive_candidate(self, candidate):
        pass

    def warn(self, message):
        exceptions.warn(message)


class Accumulator(Receiver):
    def __init__(self):
        self.test_indices = None
        self.warnings = []
        self.candidates = []

    def receive_candidate(self, candidate):
        self.candidates.append(candidate)

    def receive_context(self, context):
        self.test_indices = context.test_indices

    def warn(self, message):
        self.warnings.append(message)



import numpy as np
import pandas as pd


def select_indices(ctx, only_strings=False):
    """Returns the column indices that should be categorically encoded."""

    def should_be_encoded(col):
        if col.role != 'categorical':
            return False
        elif only_strings:
            return any(isinstance(i, str) for i in col.values)
        else:
            return True

    return ctx.matrix.column_indices_where(should_be_encoded)


def create_encoder(ctx, method, indices):
    return OneHotEncoder() if method == 'one-hot' else OrdinalEncoder()


def encode(ctx, encoder, indices):
    found = ctx.matrix.select_columns(indices)
    array = found.stack_columns()
    ctx.matrix.drop_columns_by_index(indices)

    if ctx.is_recording:
        result = encoder.fit_transform(array)
    else:
        try:
            result = encoder.transform(array)
        except Exception:
            result = _create_failed_encoding(encoder, found)
            ctx.receiver.warn('Failed to encode categorical data.')

    if isinstance(encoder, OneHotEncoder):
        formulas = _one_hot_encoded_formulas(encoder, found)
    else:
        formulas = [['encode', col] for col in found.columns]

    for i in range(result.shape[1]):
        ctx.matrix.append_column(
            values=result[:, i],
            formula=formulas[i],
            role='encoded',
        )


def _one_hot_encoded_formulas(encoder, found):
    result = []
    for series, column in zip(encoder.mapping, found.columns):
        for value in series.index:
            result.append([f'equals[{value}]', column])
    return result


def _create_failed_encoding(encoder, found):
    num_rows = len(found)
    if isinstance(encoder, OneHotEncoder):
        num_cols = sum(len(s) for s in encoder.mapping)
    else:
        num_cols = len(found.columns)
    return np.zeros((num_rows, num_cols))


class LabelEncoder:
    def __init__(self):
        self.mapping = None

    def __repr__(self):
        return 'LabelEncoder'

    @property
    def classes_(self):
        return self.mapping.index.values

    def fit_transform(self, y):
        if not isinstance(y, pd.Series):
            y = pd.Series(y)
        self.mapping = _ordinally_map_series(y)
        return self.transform(y)

    def transform(self, y):
        if not isinstance(y, pd.Series):
            y = pd.Series(y)
        y = y.map(self.mapping)
        y.fillna(0, inplace=True)
        return y.values

    def inverse_transform(self, y):
        inverse = pd.Series(data=self.mapping.index, index=self.mapping.values)
        return pd.Series(y).map(inverse).values


class OrdinalEncoder:
    def __init__(self):
        self.mapping = None

    def __repr__(self):
        return 'OrdinalEncoder'

    def fit_transform(self, X):
        X = pd.DataFrame(X)
        self.mapping = [_ordinally_map_series(X[i]) for i in X.columns]
        return self.transform(X)

    def transform(self, X):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        for i, series in enumerate(self.mapping):
            X[i] = X[i].map(series)
            X[i].fillna(0, inplace=True)
        return X.values


def _ordinally_map_series(series):
    index = [x for x in pd.unique(series) if x is not None]
    data = list(range(1, len(index) + 1))
    return pd.Series(data=data, index=index)


class OneHotEncoder:
    def __init__(self):
        self._encoder = None

    def __repr__(self):
        return 'OneHotEncoder'

    @property
    def mapping(self):
        return self._encoder.mapping

    def fit_transform(self, X):
        self._encoder = OrdinalEncoder()
        return self._elaborate(self._encoder.fit_transform(X))

    def transform(self, X):
        return self._elaborate(self._encoder.transform(X))

    def _elaborate(self, X):
        cols = []
        for i, series in enumerate(self.mapping):
            for value in series:
                cols.append(X[:, i] == value)
        return np.stack(cols, axis=1)



__version__ = '0.0.10'

from .cleaning import clean_dataset
from .context import create_context, RecordingContext, Labels
from .exceptions import Autom8Exception, Autom8Warning
from .formats import read_csv
from .inference import infer_roles
from .main import fit, run

from .matrix import (
    create_matrix,
    Column,
    Matrix,
)

from .packager import create_example_input, create_package
from .pipeline import Pipeline

from .preprocessors import (
    add_column_of_ones,
    binarize_fractions,
    binarize_signs,
    coerce_columns,
    divide_columns,
    drop_duplicate_columns,
    drop_text_columns,
    drop_weak_columns,
    encode_categories,
    encode_text,
    logarithm_columns,
    multiply_columns,
    preprocessor,
    scale_columns,
    square_columns,
    sqrt_columns,
)

from .receiver import Accumulator, Receiver



from collections import namedtuple
import functools
import itertools
import logging

import numpy as np
import sklearn.feature_extraction.text
import sklearn.preprocessing

from . import categories
from .exceptions import expected, typename


Step = namedtuple('Step', 'func, args, kwargs')


def planner(f):
    @functools.wraps(f)
    def wrapper(ctx, *a, **k):
        if not ctx.is_recording:
            raise expected('RecordingContext', typename(ctx))
        try:
            f(ctx, *a, **k)
        except Exception as exc:
            msg = f'Planning step "{f.__name__}" failed: {exc}'
            logging.getLogger('autom8').exception(msg)
            ctx.receiver.warn(msg)
    return wrapper


def preprocessor(f):
    @functools.wraps(f)
    def wrapper(ctx, *a, **k):
        f(ctx, *a, **k)
        if ctx.is_recording:
            ctx.steps.append(Step(wrapper, a, k))
    return wrapper


def playback(steps, ctx):
    for f, a, k in steps:
        try:
            f(ctx, *a, **k)
        except Exception:
            msg = f'Playback failed on step {f.__name__}'
            logging.getLogger('autom8').exception(msg)
            ctx.receiver.warn(msg)


@preprocessor
def add_column_of_ones(ctx):
    """Adds a column containing only the number 1, to help some estimators."""
    columns = ctx.matrix.columns
    num_rows = len(columns[0]) if columns else 0
    ctx.matrix.append_column(
        values=np.full(num_rows, 1),
        formula=['constant(1)'],
        role='numerical',
    )


@planner
def binarize_fractions(ctx):
    indices = _original_numerical_indices(ctx)
    if indices:
        _binarize_fractions(ctx, indices)


@preprocessor
def _binarize_fractions(ctx, indices):
    """Indicates when a column contains a fractional value."""
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for col in found.columns:
        ctx.matrix.append_column(
            values=np.absolute(col.values) < 1,
            formula=['is-fraction', col],
            role='encoded',
        )


@planner
def binarize_signs(ctx):
    indices = _original_numerical_indices(ctx)
    if indices:
        _binarize_signs(ctx, indices)


@preprocessor
def _binarize_signs(ctx, indices):
    """Indicates when a column contains a positive value."""
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for col in found.columns:
        ctx.matrix.append_column(
            values=col.values > 0,
            formula=['is-positive', col],
            role='encoded',
        )


@planner
def coerce_columns(ctx):
    dtypes = [col.dtype for col in ctx.matrix.columns]
    _coerce_columns(ctx, dtypes)


@preprocessor
def _coerce_columns(ctx, dtypes):
    """Forces each column to have the expected type."""

    for col, dtype in zip(ctx.matrix.columns, dtypes):
        if col.dtype == dtype or dtype == object:
            continue
        try:
            col.values = col.values.astype(dtype)
        except Exception:
            ctx.receiver.warn(
                f'Failed to convert column "{col.name}" to {dtype}'
            )


@planner
def divide_columns(ctx):
    numerators = _original_numerical_indices(ctx)
    # Find the indices of the original numerical columns that don't contain
    # any 0 values.
    denominators = _original_numerical_indices(ctx, lambda c: np.all(c.values != 0))

    if numerators and denominators:
        _divide_columns(ctx, numerators, denominators)


@preprocessor
def _divide_columns(ctx, numerators, denominators):
    """
    Divides pairs of numerical columns, as long as the divisor column
    does not contain any 0 values.
    """
    xs = ctx.matrix.select_columns(numerators)
    ys = ctx.matrix.select_columns(denominators)
    xs.coerce(float)
    ys.coerce(float)
    for col_x, col_y in itertools.product(xs.columns, ys.columns):
        x, y = col_x.values, col_y.values
        if not np.array_equal(x, y):
            # Use 0 if we see a denominator that equals 0.
            ctx.matrix.append_column(
                values=np.divide(x, y, out=np.zeros_like(x), where=y != 0),
                formula=['divide', col_x, col_y],
                role='numerical',
            )


@planner
def drop_duplicate_columns(ctx):
    duplicates = []
    visited = set()
    for index, col in enumerate(ctx.matrix.columns):
        signature = col.values.tobytes()
        if signature in visited:
            duplicates.append(index)
        else:
            visited.add(signature)
    if duplicates:
        _drop_duplicate_columns(ctx, duplicates)


@preprocessor
def _drop_duplicate_columns(ctx, duplicates):
    """Removes duplicate columns from your dataset."""
    ctx.matrix.drop_columns_by_index(duplicates)


@planner
def drop_weak_columns(ctx, feature_selector=None):
    import sklearn.feature_selection as fs

    if feature_selector is None and ctx.is_regression:
        feature_selector = fs.SelectFwe(fs.f_regression)

    if feature_selector is None and ctx.is_classification:
        feature_selector = fs.SelectFwe(fs.f_classif)

    X, y = ctx.training_data()
    X = X._float_array()

    feature_selector.fit(X, y)
    weak_cols = np.where(np.invert(feature_selector.get_support()))[0]

    # For now, ignore the selector if it wants to drop every column.
    if 0 < len(weak_cols) < len(ctx.matrix.columns):
        _drop_weak_columns(ctx, weak_cols.tolist())


@preprocessor
def _drop_weak_columns(ctx, indices):
    """Selects the best columns in your dataset."""
    ctx.matrix.drop_columns_by_index(indices)


@planner
def encode_categories(ctx, method='ordinal', only_strings=False):
    if method not in {'one-hot', 'ordinal'}:
        raise expected('one-hot or ordinal', method)

    indices = categories.select_indices(ctx, only_strings=only_strings)

    if indices:
        encoder = categories.create_encoder(ctx, method, indices)
        _encode_categories(ctx, encoder, indices)


@preprocessor
def _encode_categories(ctx, encoder, indices):
    """Turns categorical features into columns of numbers."""
    categories.encode(ctx, encoder, indices)


@planner
def encode_text(ctx):
    indices = ctx.matrix.column_indices_where(lambda col: col.role == 'textual')
    if indices:
        encoder = sklearn.feature_extraction.text.TfidfVectorizer(
            stop_words='english',
            min_df=5,
            max_features=2500,
            sublinear_tf=True
        )
        _encode_text(ctx, encoder, indices)


@preprocessor
def _encode_text(ctx, encoder, indices):
    """Turns textual features into vectors of numbers."""
    found = ctx.matrix.select_columns(indices)
    ctx.matrix.drop_columns_by_index(indices)

    # Combine textual columns into single column.
    found.coerce(str)
    combined = [' '.join(text) for text in found.stack_columns()]

    if ctx.is_recording:
        result = encoder.fit_transform(combined).toarray()
    else:
        # TODO: Append the appropriate number of 0-columns when this fails.
        result = encoder.transform(combined).toarray()

    for i in range(result.shape[1]):
        ctx.matrix.append_column(
            values=result[:, i],
            formula=['encode-text'] + found.columns,
            role='encoded',
        )


@planner
def drop_text_columns(ctx):
    indices = ctx.matrix.column_indices_where(lambda col: col.role == 'textual')
    if indices:
        _drop_text_columns(ctx, indices)


@preprocessor
def _drop_text_columns(ctx, indices):
    """Removes text data from your dataset."""
    ctx.matrix.drop_columns_by_index(indices)


@planner
def logarithm_columns(ctx):
    # Find the indices of the original numerical columns that only contain
    # positive values.
    indices = _original_numerical_indices(ctx, lambda c: np.all(c.values > 0))
    if indices:
        _logarithm_columns(ctx, indices)


@preprocessor
def _logarithm_columns(ctx, indices):
    """
    Computes the natural log each numerical column, as long as the column
    only contains positive numbers.
    """
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for col in found.columns:
        v = col.values
        ctx.matrix.append_column(
            values=np.log(v, out=np.zeros_like(v), where=v > 0),
            formula=['log', col],
            role='numerical',
        )


@planner
def multiply_columns(ctx):
    indices = _original_numerical_indices(ctx)
    if indices:
        _multiply_columns(ctx, indices)


@preprocessor
def _multiply_columns(ctx, indices):
    """Multiplies pairs of numerical columns."""
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for x, y in itertools.combinations(found.columns, 2):
        ctx.matrix.append_column(
            values=x.values * y.values,
            formula=['multiply', x, y],
            role='numerical',
        )


@planner
def scale_columns(ctx, scaler=None):
    indices = ctx.matrix.column_indices_where(lambda col: col.is_numerical)
    if indices and scaler is None:
        scaler = sklearn.preprocessing.RobustScaler(quantile_range=(5, 95))

    if indices:
        _scale_columns(ctx, scaler, indices)


@preprocessor
def _scale_columns(ctx, scaler, indices):
    """Standardizes numerical columns using different scaling strategies."""
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)

    array = found.stack_columns()
    ctx.matrix.drop_columns_by_index(indices)

    if ctx.is_recording:
        result = scaler.fit_transform(array)
    else:
        try:
            result = scaler.transform(array)
        except Exception:
            result = np.zeros_like(array)

    for i in range(result.shape[1]):
        ctx.matrix.append_column(
            values=result[:, i],
            formula=['scale', found.columns[i]],
            role='numerical',
        )


@planner
def square_columns(ctx):
    indices = _original_numerical_indices(ctx)
    if indices:
        _square_columns(ctx, indices)


@preprocessor
def _square_columns(ctx, indices):
    """Multiplies each numerical column by itself."""
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for col in found.columns:
        ctx.matrix.append_column(
            values=col.values * col.values,
            formula=['square', col],
            role='numerical',
        )


@planner
def sqrt_columns(ctx):
    # Find the indices of the original numerical columns that only contain
    # values greater than or equal to zero.
    indices = _original_numerical_indices(ctx, lambda c: np.all(c.values >= 0))
    if indices:
        _sqrt_columns(ctx, indices)


@preprocessor
def _sqrt_columns(ctx, indices):
    """
    Computes the square root of each numerical column, as long as the column
    does not contain any negative numbers.
    """
    found = ctx.matrix.select_columns(indices)
    found.coerce(float)
    for col in found.columns:
        v = col.values
        ctx.matrix.append_column(
            values=np.sqrt(v, out=np.zeros_like(v), where=v >= 0),
            formula=['square-root', col],
            role='numerical',
        )


def _original_numerical_indices(ctx, where=None):
    if where is None:
        where = lambda col: True
    return ctx.matrix.column_indices_where(lambda col:
        col.is_numerical and col.is_original and where(col))



import csv
from functools import reduce
from io import StringIO

import chardet


def decode_csv(payload):
    """Takes the contents of a CSV file and returns a list of rows.

    Parameters:
        payload (bytes or str): The contents of a CSV file.

    Returns:
        list[row]: A list of rows. Each row is a list of strings and numbers.
    """

    def _convert(cell):
        try:
            return parse_number(cell)
        except Exception:
            return cell

    if isinstance(payload, bytes):
        payload = payload.decode(chardet.detect(payload)['encoding'])

    # SHOULD: Make this work for quoted values with newlines in the first row.
    dialect = csv.Sniffer().sniff(payload.split('\n')[0])
    reader = csv.reader(StringIO(payload), dialect)
    return [[_convert(i) for i in row] for row in reader]


def drop_empty_rows(rows):
    """Takes a list of rows and returns a new list without any blank rows.

    autom8 considers a row to be blank if it's totally empty, or if each item
    in the row is None, the empty string, or a string that only contain spaces.

    >>> drop_empty_rows([])
    []

    >>> drop_empty_rows([[], [1, 2, 3], [], [4, 5, 6], []])
    [[1, 2, 3], [4, 5, 6]]

    >>> drop_empty_rows([[9, 8, 7], [' ', '\t', ''], [6, 5, 4], [None, ' ']])
    [[9, 8, 7], [6, 5, 4]]
    """

    def _is_blank(obj):
        if isinstance(obj, str):
            return obj == '' or obj.isspace()
        else:
            return obj is None

    return [row for row in rows if any(not _is_blank(i) for i in row)]


def encode_csv(dataset):
    """Takes a list of rows and returns it as a CSV string.

    Parameters:
        dataset (list[row]): A list of rows. Each row is a list of primitive
            values.

    Returns:
        str: A CSV representation of the dataset.

    >>> encode_csv([[1, 2, 3], [4, 5, 6]])
    '1,2,3\\r\\n4,5,6\\r\\n'

    >>> encode_csv([['a, b, c', 'xyz', None, 1.0]])
    '"a, b, c",xyz,,1.0\\r\\n'

    >>> encode_csv([['he said, "well, ok" and then left', 'foo', 'bar']])
    '"he said, ""well, ok"" and then left",foo,bar\\r\\n'
    """

    result = StringIO()
    writer = csv.writer(result, quoting=csv.QUOTE_MINIMAL)
    for row in dataset:
        writer.writerow(row)
    return result.getvalue()


def excel_column_index(name):
    """Takes the Excel-style name of a column and returns its 0-based index.

    >>> excel_column_index('A')
    0
    >>> excel_column_index('B')
    1
    >>> excel_column_index('Z')
    25
    >>> excel_column_index('AA')
    26
    >>> excel_column_index('AZ')
    51
    >>> excel_column_index('BC')
    54
    """

    A = ord('A')
    return reduce(lambda acc, c: acc * 26 + (ord(c) - A + 1), name, 0) - 1


def excel_column_name(num):
    """Takes the 0-based index of a column, and returns its Excel-style name.

    >>> excel_column_name(0)
    'A'
    >>> excel_column_name(1)
    'B'
    >>> excel_column_name(23)
    'X'
    >>> excel_column_name(26)
    'AA'
    >>> excel_column_name(57)
    'BF'
    """

    A = ord('A')
    result = []
    remaining = num + 1
    while remaining > 0:
        modulo = (remaining - 1) % 26
        result.append(chr(A + modulo))
        remaining = (remaining - modulo) // 26
    return ''.join(reversed(result))


def parse_number(string):
    """Takes a string and attempts to convert it to a number.

    This function simply removes commas and underscores from the string, and
    then tries to convert it to an int or a float.

    >>> parse_number('10')
    10

    >>> parse_number('1.4')
    1.4

    >>> parse_number('12,345')
    12345

    >>> parse_number('987_654_321')
    987654321

    >>> parse_number('0xff')
    255

    >>> parse_number('0XFE')
    254

    >>> parse_number({})
    Traceback (most recent call last):
        ...
    TypeError: parse_number() argument must be a string

    >>> parse_number('1.2.3')
    Traceback (most recent call last):
        ...
    Exception: invalid number literal: '1.2.3'
    """

    if not isinstance(string, str):
        raise TypeError('parse_number() argument must be a string')

    string = string.replace(',', '').replace('_', '')

    try:
        return int(string)
    except Exception:
        pass

    try:
        return float(string)
    except Exception:
        pass

    try:
        if string.startswith(('0x', '0X')):
            return int(string, 16)
    except Exception:
        pass

    raise Exception(f'invalid number literal: {repr(string)}')


def read_csv(path):
    """Reads the CSV file at the indicated path and returns a list of rows.

    Parameters:
        path (str): The path to a CSV file.

    Returns:
        list[row]: A list of rows. Each row is a list of strings and numbers.
    """

    with open(path, 'rb') as f:
        return decode_csv(f.read())



import re
import numpy as np

from .docstrings import render_docstring
from .exceptions import expected, typename
from .formats import (
    drop_empty_rows,
    excel_column_index,
    excel_column_name,
    parse_number,
)
from .receiver import Receiver


@render_docstring
def create_matrix(dataset, column_names=None, column_roles=None, receiver=None):
    """Returns a new Matrix object from the provided dataset.

    Parameters:
        $dataset_parameters
        $receiver_parameter
    """

    if receiver is None:
        receiver = Receiver()

    matrix = _create_matrix(dataset, column_names, column_roles, receiver)

    # Warn the user if the column names are not unique.
    names = matrix.column_names
    if len(names) != len(set(names)):
        receiver.warn(f'Column names are not unique in {repr(names)}')

    return matrix


def _create_matrix(dataset, names, roles, receiver):
    if not isinstance(dataset, (list, tuple, np.ndarray, Matrix)):
        raise expected('list, tuple, numpy array, or Matrix', typename(dataset))

    if isinstance(dataset, Matrix):
        return _copy_and_update_matrix(dataset, names, roles)

    # Drop empty rows.
    rows = drop_empty_rows(dataset)
    num_dropped = len(dataset) - len(rows)

    # Warn the users if we dropped some rows.
    if num_dropped:
        suffix = '' if num_dropped == 1 else 's'
        receiver.warn(f'Dropped {num_dropped} empty row{suffix} from dataset.')

    # Figure out how many columns we need.
    mincols = min(len(row) for row in rows) if rows else 0
    maxcols = max(len(row) for row in rows) if rows else 0

    # Warn the user if we're dropping any extra columns.
    if mincols < maxcols:
        num_extra = maxcols - mincols
        suffix1 = '' if num_extra == 1 else 's'
        suffix2 = '' if mincols == 1 else 's'
        receiver.warn(
            f'Dropped {num_extra} extra column{suffix1} from dataset.'
            f' Keeping first {mincols} column{suffix2}.'
            ' To avoid this behavior, ensure that each row in the dataset has'
            ' the same number of columns.'
        )

    def make(index):
        values = create_array([row[index] for row in rows])
        formula = excel_column_name(index)
        return Column(values=values, formula=formula, role=None, is_original=True)

    matrix = Matrix([make(i) for i in range(mincols)])
    _name_columns(matrix, names)
    _update_roles(matrix, roles)
    return matrix


class Matrix:
    """Represents a matrix of features.

    Attributes:
        columns (list[Column]): A list of the matrix's columns.
    """

    def __init__(self, columns):
        assert isinstance(columns, list)
        assert all(isinstance(i, Column) for i in columns)
        self.columns = columns

    @property
    def column_names(self):
        return [col.name for col in self.columns]

    @property
    def formulas(self):
        return [col.formula for col in self.columns]

    def __len__(self):
        return len(self.columns[0]) if self.columns else 0

    def copy(self):
        return Matrix([i.copy() for i in self.columns])

    def tolist(self):
        cols = [c.values.astype(object) for c in self.columns]
        return np.column_stack(cols).tolist()

    def stack_columns(self):
        return np.column_stack([col.values for col in self.columns])

    def to_array(self):
        if len(self.columns) == 1:
            return self.columns[0].values
        else:
            raise expected('matrix with only one column', len(self.columns))

    def append_column(self, values, formula, role, is_original=False):
        if not hasattr(values, 'shape'):
            raise expected('numpy array', type(values))

        if len(values.shape) != 1:
            raise expected('array of values', values.shape)

        col = Column(values, formula=formula, role=role, is_original=is_original)
        self.columns.append(col)

    def drop_columns_by_index(self, indices):
        if isinstance(indices, int):
            indices = [indices]

        self.columns = [col for i, col in enumerate(self.columns) if i not in indices]

    def select_rows(self, indices):
        return Matrix([col.select_rows(indices) for col in self.columns])

    def exclude_rows(self, indices):
        return Matrix([col.exclude_rows(indices) for col in self.columns])

    def select_columns(self, indices):
        return Matrix([self.columns[i].copy() for i in indices])

    def exclude_columns(self, indices):
        return Matrix([col.copy()
            for i, col in enumerate(self.columns)
                if i not in indices])

    def select_columns_by_name(self, names):
        # If the names match exactly, then return the matrix as-is.
        if self.column_names == names:
            return self

        # If this matrix has the requested columns, then make a new matrix with them.
        if set(self.column_names).issuperset(names):
            lookup = {col.name: col for col in self.columns}
            return Matrix([lookup[n].copy() for n in names])

        raise expected(f'column names to be in {self.column_names}', names)

    def column_indices_where(self, predicate):
        return [i for i, col in enumerate(self.columns) if predicate(col)]

    def coerce(self, to_type):
        for col in self.columns:
            col.coerce(to_type)

    def _float_array(self):
        """Warning: This method mutates the matrix object."""
        self.coerce(float)
        return self.stack_columns()


class Column:
    def __init__(self, values, formula, role, is_original):
        assert len(values.shape) == 1
        assert isinstance(formula, (str, list))
        assert role is None or isinstance(role, str)
        assert isinstance(is_original, bool)
        self.values = values

        if isinstance(formula, str):
            self.formula = formula
        else:
            self.formula = [getattr(col, 'formula', col) for col in formula]

        self._role = None
        self.role = role
        self.is_original = is_original

    def __len__(self):
        return len(self.values)

    @property
    def name(self):
        if isinstance(self.formula, str):
            return self.formula
        else:
            # TODO: Pretty-print the formula. Maybe put "=" in front.
            return repr(self.formula)

    @property
    def dtype(self):
        return self.values.dtype

    @property
    def is_numerical(self):
        return self.role == 'numerical'

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, role):
        valid_roles = {None, 'categorical', 'encoded', 'numerical', 'textual'}

        if role not in valid_roles:
            raise expected(f'role in {valid_roles}', repr(role))

        self._role = role

    def copy(self):
        values = np.copy(self.values)
        return self.copy_with(values)

    def copy_with(self, values, role='__copy__'):
        assert len(values.shape) == 1

        # Use a magic sentinel instead of None, since None is a valid role.
        if role == '__copy__':
            role = self.role

        return Column(
            values=values,
            formula=self.formula,
            role=role,
            is_original=self.is_original,
        )

    def coerce(self, to_type):
        expected = (bool, float, int, object, str)
        if to_type not in expected:
            raise TypeError(f'coerce() argument must be one of: {expected}')

        if to_type == bool:
            self._coerce(bool, False, bool)

        elif to_type in [float, int]:
            self.coerce_values_to_numbers(to_type)

        elif to_type == object:
            self.values = self.values.astype(object, copy=False)

        elif to_type == str:
            self.coerce_values_to_strings()

        else:
            assert False

    def coerce_values_to_numbers(self, to_type=float):
        def conv(x):
            try:
                return to_type(parse_number(x))
            except Exception:
                return to_type(x)

        self._coerce(to_type, to_type(0), conv)

    def coerce_values_to_strings(self):
        def conv(x):
            try:
                return str(x)
            except Exception:
                return ''

        if not all(isinstance(x, str) for x in self.values):
            self.values = create_array([conv(x) for x in self.values])

    def _coerce(self, to_type, default, coerce_func):
        try:
            self.values = self.values.astype(to_type, copy=False)
            return
        except Exception:
            pass

        def conv(x):
            if isinstance(x, to_type):
                return x
            try:
                return coerce_func(x)
            except Exception:
                return default

        self.values = np.array([conv(x) for x in self.values], dtype=to_type)

    def select_rows(self, indices):
        return self.copy_with(self.values[indices])

    def exclude_rows(self, indices):
        return self.copy_with(np.delete(self.values, indices))

    def root_columns(self):
        roots = set()
        stack = [self.formula]
        while stack:
            f = stack.pop()
            if isinstance(f, str):
                roots.add(f)
            else:
                assert isinstance(f, (list, tuple))
                assert isinstance(f[0], str)
                stack.extend(f[1:])
        return roots


def _copy_and_update_matrix(matrix, names, roles):
    matrix = matrix.copy()

    if names is not None:
        _name_columns(matrix, names)

    if roles is not None:
        _update_roles(matrix, roles)

    return matrix


def _name_columns(matrix, names):
    # If the column names are included in the dataset, then remove the first
    # value from each column and use that at its name.
    if isinstance(names, str) and names == 'included':
        _extract_column_names(matrix)
        return

    # If the column names are missing from the dataset, then assign a new,
    # unique name to each column.
    if isinstance(names, str) and names == 'missing':
        _generate_column_names(matrix)
        return

    # If the column names are provided separately, then assign a name to each
    # column. Raise an exception if we don't have the right number of names.
    if isinstance(names, (list, tuple)):
        _update_column_names(matrix, names)
        return

    # Make sure that names is either 'unknown' or None.
    if names not in {'unknown', None}:
        valid_codes = {'included', 'missing', 'unknown', None}
        raise expected(
            f'column names to be a list of strings, or one of {valid_codes}',
            repr(names)
        )

    # Figure out whether the column names are included or missing, and then recur.
    are_included = _includes_column_names(matrix)
    _name_columns(matrix, 'included' if are_included else 'missing')


def _includes_column_names(matrix):
    # If the matrix is empty, then the column names must be missing.
    if len(matrix) == 0:
        return False

    # If the first row contains all strings, then assume those are the
    # column names.
    if all(isinstance(col.values[0], str) for col in matrix.columns):
        return True

    # If any column looks like it starts with a name, then assume that this
    # matrix must include the column names.
    for col in matrix.columns:
        vals = col.values
        has_name = isinstance(vals[0], str)
        has_rest = len(vals) > 1
        has_nums = any(i is not None and not isinstance(i, str) for i in vals[1:])
        no_strs = all(not isinstance(i, str) or i == '' for i in vals[1:])
        if has_name and has_rest and has_nums and no_strs:
            return True

    # Well, it doesn't look like this matrix contains the column names.
    return False


def _extract_column_names(matrix):
    if len(matrix) == 0:
        raise expected('nonempty dataset', 'empty dataset')

    for col in matrix.columns:
        # Use the first value, even if it's not a string.
        col.formula = str(col.values[0])

        # Let numpy infer a (potentially) new dtype.
        col.values = create_array(col.values[1:].tolist())


def _generate_column_names(matrix):
    for i, col in enumerate(matrix.columns):
        col.formula = excel_column_name(i)


def _update_column_names(matrix, names):
    assert isinstance(names, (list, tuple))

    if len(matrix.columns) != len(names):
        raise expected(f'{len(matrix.columns)} column names', len(names))

    if not all(isinstance(i, str) for i in names):
        raise expected('column names to be a list of strings', repr(names))

    for col, name in zip(matrix.columns, names):
        col.formula = name


def _update_roles(matrix, roles):
    if roles is None:
        return

    if isinstance(roles, (list, tuple)):
        _assign_roles(matrix, roles)
        return

    if not isinstance(roles, dict):
        raise expected('column roles to be a list or a dict', repr(roles))

    colmap = {col.name: col for col in matrix.columns}

    def lookup(key):
        if isinstance(key, str) and key not in colmap:
            names = [col.name for col in matrix.columns]
            raise expected(f'column to be one of {names}', repr(key))

        if isinstance(key, str):
            return colmap[key]

        try:
            return matrix.columns[key]
        except Exception:
            pass

        raise expected('valid column key', repr(key))

    for key, role in roles.items():
        col = lookup(key)
        col.role = role


def _assign_roles(matrix, roles):
    if len(matrix.columns) != len(roles):
        raise expected(f'{len(matrix.columns)} column roles', len(roles))

    for col, role in zip(matrix.columns, roles):
        col.role = role


def create_array(values):
    has_str = any(isinstance(i, str) for i in values)
    return np.array(values, dtype=object if has_str else None)



from .preprocessors import planner, preprocessor


@planner
def infer_roles(ctx):
    """Infers the role of each column.

    autom8 uses a handful of roles: `categorical`, `encoded`, `numerical`, and
    `textual`.

    autom8 infers that a column contains categorical data when the column
    contains fewer than 50 unique values, and when less than 25% of the values
    are unique.

    autom8 infers that a column contains encoded data when the column only
    contains boolean values.

    Otherwise, if a column contains strings, then autom8 infers that it
    contains textual data. And if it contains numbers, then autom8 infers that
    it contains numerical data.
    """

    roles = [_infer_role(col, ctx.receiver) for col in ctx.matrix.columns]
    _set_roles(ctx, roles)


@preprocessor
def _set_roles(ctx, roles):
    """Records the type of each column."""
    for col, role in zip(ctx.matrix.columns, roles):
        col.role = role


def _infer_role(col, receiver):
    inferred = _infer_role_from_values(col.values)
    return _merge_roles(col, inferred, receiver)


def _infer_role_from_values(values):
    # If it's a column of booleans, then just say that it's already encoded.
    if values.dtype == bool:
        return 'encoded'

    # And if it's a column of floats, then say that it's numerical.
    if values.dtype == float:
        return 'numerical'

    num_unique = len(set(values))
    ratio = num_unique / len(values)
    is_categorical = ratio <= 0.25 and num_unique < 50
    is_all_strings = all(isinstance(e, str) for e in values)

    if is_categorical:
        return 'categorical'
    elif is_all_strings:
        return 'textual'
    else:
        return 'numerical'


def _merge_roles(col, new_role, receiver):
    # Let's use short names in here.
    old, new = col.role, new_role

    # Complain if one is "numerical" and the other is "textual".
    if (old, new) in {('numerical', 'textual'), ('textual', 'numerical')}:
        receiver.warn(
            f'Found a {new} column declared as a {old} column: {col.name}'
        )
        return new

    # If either one is None, return the other one.
    if new is None or old is None:
        return new or old

    # If the inferencer thinks its encoded, and the user says its categorical,
    # then we can stick with "encoded".
    if new == 'encoded' and old == 'categorical':
        return 'encoded'

    # In any other situation, just trust whatever the old value says.
    return old



import sklearn.linear_model
import sklearn.naive_bayes
import sklearn.neighbors
import sklearn.neural_network
import sklearn.svm
import sklearn.tree

try:
    import lightgbm
except ImportError:
    lightgbm = None

try:
    import xgboost
except ImportError:
    xgboost = None

from .cleaning import clean_dataset
from .docstrings import render_docstring
from .inference import infer_roles
from . import preprocessors as then
from .context import create_context
from .selector import create_selector


@render_docstring
def fit(*args, **kwargs):
    """Runs autom8 with the provided settings and returns the best pipeline.

    Parameters:
        $common_context_parameters
        $selector_parameters

    Returns:
        autom8.Pipeline: The best pipeline, as indicated by the `selector`
        parameter.
    """

    if 'receiver' in kwargs:
        raise TypeError("fit() got an unexpected keyword argument 'receiver'")

    selector = create_selector(kwargs.pop('selector', None))
    run(*args, **kwargs, receiver=selector)
    return None if selector.best is None else selector.best.pipeline


@render_docstring
def run(*args, **kwargs):
    """Runs autom8 with the provided settings.

    After creating the `RecordingContext` object, autom8 passes the context to
    the receiver's `receive_context` method. The allows the receive to copy any
    information that it may need later, like the context's `test_indices`
    attribute, for example.

    As autom8 creates candidate pipelines, it passes each candidate to the
    receiver's `receive_candidate` method. At that point, the receiver is free
    to do whatever it wants with the candidate.

    Parameters:
        $all_context_parameters
    """

    ctx = create_context(*args, **kwargs)
    ctx.receiver.receive_context(ctx)

    clean_dataset(ctx)
    infer_roles(ctx)

    with ctx.sandbox():
        then.encode_categories(ctx, method='ordinal', only_strings=True)
        ctx.receiver.receive_cleaned_context(ctx)

    then.coerce_columns(ctx)
    then_ignore_text_columns(ctx)
    then.encode_text(ctx)

    with ctx.sandbox():
        then.encode_categories(ctx, method='ordinal', only_strings=True)
        then_fit_estimators(ctx, boosted_trees=True, classical=False)

        then_finish_up_preprocessing(ctx)
        then_fit_estimators(ctx, boosted_trees=True, classical=True)

    with ctx.sandbox():
        then.encode_categories(ctx, method='ordinal', only_strings=True)
        then.multiply_columns(ctx)
        then.add_column_of_ones(ctx)
        then_fit_estimators(ctx, boosted_trees=True, classical=False)

    with ctx.sandbox():
        # This time around, use one-hot encoding and skip the boosted trees.
        then.encode_categories(ctx, method='one-hot', only_strings=False)

        # Try not to engineer more columns than the number of rows in the
        # test dataset.
        max_num_cols = len(ctx.matrix.columns) + len(ctx.test_indices)
        has_room = lambda: len(ctx.matrix.columns) < max_num_cols

        if has_room(): then.multiply_columns(ctx)
        if has_room(): then.square_columns(ctx)
        if has_room(): then.divide_columns(ctx)
        if has_room(): then.binarize_signs(ctx)
        if has_room(): then.logarithm_columns(ctx)
        if has_room(): then.sqrt_columns(ctx)
        if has_room(): then.binarize_fractions(ctx)

        then_finish_up_preprocessing(ctx)
        then_fit_estimators(ctx, boosted_trees=False, classical=True)


def then_ignore_text_columns(ctx):
    has_text = any(col.role == 'textual' for col in ctx.matrix.columns)
    all_text = all(col.role == 'textual' for col in ctx.matrix.columns)

    if has_text and not all_text:
        with ctx.sandbox():
            then.drop_text_columns(ctx)
            # Just try the boosted trees for now.
            then.encode_categories(ctx, method='ordinal', only_strings=True)
            then_fit_estimators(ctx, boosted_trees=True, classical=False)


def then_finish_up_preprocessing(ctx):
    # TODO: Search for a good scaler.
    then.scale_columns(ctx)
    then.drop_weak_columns(ctx)
    then.add_column_of_ones(ctx)
    then.drop_duplicate_columns(ctx)


def then_fit_estimators(ctx, **flags):
    with ctx.parallel():
        if ctx.is_regression:
            fit_regressors(ctx, **flags)

        if ctx.is_classification:
            fit_classifiers(ctx, **flags)


def fit_regressors(ctx, boosted_trees=True, classical=True):
    random_state = ctx.random_state_kw

    if boosted_trees and lightgbm is not None:
        ctx << lightgbm.LGBMRegressor(n_jobs=n_jobs(ctx), **random_state)

    if boosted_trees and xgboost is not None:
        ctx << xgboost.XGBRegressor(n_jobs=n_jobs(ctx), **random_state)

    # Always fit a linear regression.
    ctx << sklearn.linear_model.LinearRegression()

    # Only run SGD when the dataset is big (and when `classical` is True).
    if classical and is_big(ctx):
        ctx << sklearn.linear_model.SGDRegressor(
            learning_rate='constant',
            eta0=0.00001,
            max_iter=1000,
            tol=1e-3,
            **random_state,
        )

    # Non-SGD estimators struggle to converge when there are too many rows.
    if classical and not is_big(ctx):
        ctx << sklearn.tree.DecisionTreeRegressor(**random_state)

        ctx << sklearn.linear_model.ElasticNetCV(
            l1_ratio=0.0,
            alphas=[0.1, 1.0, 10.0],
            **random_state,
        )

        ctx << sklearn.linear_model.LassoCV(**random_state)
        ctx << sklearn.linear_model.ElasticNetCV(**random_state)
        ctx << sklearn.svm.LinearSVR(**random_state)

        # TODO: Consider only running this on appropriately-sized datasets.
        ctx << sklearn.neural_network.MLPRegressor(
            hidden_layer_sizes=(10, 10),
            **random_state,
        )


def fit_classifiers(ctx, boosted_trees=True, classical=True):
    random_state = ctx.random_state_kw

    if boosted_trees and lightgbm is not None:
        ctx << lightgbm.LGBMClassifier(
            class_weight='balanced', n_jobs=n_jobs(ctx), **random_state,
        )

    if boosted_trees and xgboost is not None:
        ctx << xgboost.XGBClassifier(
            class_weight='balanced', n_jobs=n_jobs(ctx), **random_state,
        )

    # Skip the decision tree when the dataset is large.
    if classical and not is_big(ctx):
        ctx << sklearn.tree.DecisionTreeClassifier(**random_state)

    if classical:
        ctx << sklearn.neighbors.KNeighborsClassifier()
        ctx << sklearn.naive_bayes.BernoulliNB(alpha=0.01)

        ctx << sklearn.linear_model.SGDClassifier(
            class_weight='balanced', max_iter=1000, tol=1e-3, **random_state,
        )

        ctx << sklearn.neural_network.MLPClassifier(
            hidden_layer_sizes=(10, 10), **random_state,
        )


def is_big(ctx):
    # TODO: Implement a smarter heuristic. 90k is a rough heuristic that does
    # not take into account the quality or quantity of features.
    return len(ctx.matrix) >= 90000


def n_jobs(ctx):
    return -1 if ctx.allow_multicore else 1



"""Utility module that allows multiple functions to share documentation."""

from string import Template

_render = lambda docstring: _strip(Template(docstring).substitute(globals()))
_strip = lambda section: section.strip() + '\n'


dataset_parameters = _strip("""
        dataset (list or tuple or numpy.ndarray or autom8.Matrix): The dataset.

            The `dataset` parameter may be a list of rows, a tuple of rows,
            a `numpy.ndarray` object of rows, or an autom8.Matrix object.

            Each row should contain booleans, numbers, strings, or None values.
            Columns with other kinds of values may be dropped by autom8.

            If the dataset contains any empty rows, or rows that only contain
            blank values, then autom8 will remove those rows from the matrix.

            If some rows contain more values than other rows, then autom8 will
            drop the extra columns, so that each row contains the same number
            of columns.

        column_names (list[str] or str or None): The column names for this
            dataset. Defaults to None.

            If `column_names` is a list of strings, then autom8 will use these
            strings as the names of the dataset's columns.

            If `column_names` is the string `included`, then autom8 will treat
            the first value in each column as the name of that column.

            If `column_names` is the string `missing`, then autom8 will
            generate a name for each column. When generating column names,
            autom8 follows the Excel-style convention of A, B, C, ..., AA, AB,
            ..., BA, BB, etc.

            If `column_names` is the string `unknown` or the value `None`, then
            autom8 will attempt to guess whether it should treat the first
            value of each column as its name, or generate a name for each
            column.

        column_roles (list or dict or None): The roles of some or all of
            the dataset's columns. Defaults to None.

            If `column_roles` is a list of roles, then autom8 will assign each
            role to the corresponding column. (For example, autom8 will assign
            the first role in the list to the first column, the second role to
            the second column, and so on.)

            If `column_roles` is a dict, then the keys may be strings or
            integers. Each string must be the name of a column, and each number
            must be a valid column index.

            The dict does not need to contain a key for each column.

            For each entry in the dict, autom8 looks up the column indicated by
            the entry's key, and assigns it the role indicated by the entry's
            value.

            Each role must be one of these values:

            - `categorical` -- Indicates the column contains categorical data.
            - `encoded` -- Indicates the column's data is already encoded and
              should be treated as opaque data.
            - `numerical` -- Indicates the column contains numerical values.
            - `textual` -- Indicates the column contains text.
            - None -- Indicates that autom8 should attempt to guess the role
              of the column's data.
""")


problem_parameters = _strip("""
        target_column (str or int or None): The column that you want your model
            to predict. Defaults to None.

            If `target_column` is a string, then it must be the name of a
            column in the dataset.

            If `target_column` is an integer, then it must be the index of a
            column in the dataset, or `-1` for the last column.

            If `target_column` is None or `-1`, then autom8 uses the dataset's
            last column as the target column.

        problem_type (str or None): The string `regression` or `classification`
            or None. Defaults to None.

            If the `problem_type` is None, then autom8 will use the role of the
            target column to determine the problem type. If the target column's
            role is `numerical`, then autom8 will use `regression`. Otherwise,
            it will use `classification`.

        test_ratio (float or None): A float between 0 and 1, or None. Defaults
            to None.

            The `test_ratio` indicates how many rows autom8 should use in its
            test dataset.

            If the `test_ratio` is None, autom8 uses a reasonable default
            (currently `0.2`).

        random_state (int or None): The seed used by each estimator's
            random number generator. Defaults to None.

        allow_multicore (bool): Allow estimators to use multiple cores.
            Defaults to True.

        executor_class (class or callable or None): An optional class for
            scheduling estimators to run. Defaults to None.

            The `executor_class` may be a class or a callable. `autom8` calls
            this object when it wants to run multiple estimators in parallel.
            This should return a new object with a few methods:

            - `fit(context, estimator)`: Schedules `context.fit(estimator)` to
              run at some point in the future (which may be right now).
            - `submit(func, *args, **kwargs)`: Schedules `func(*args, **kwargs)`
              to run at some point in the future (which may be right now).
            - `shutdown(self, wait=True)`: Shuts down the executor. If `wait`
              is True, then blocks until all tasks are complete.
""")


receiver_parameter = _strip("""
        receiver (autom8.Receiver or None): The optional Receiver object, for
            receiving out-of-band data from autom8. Defaults to None.

            If the receiver is `None`, autom8 creates a new `autom8.Receiver`
            object that ignores contexts and candidates, and passes any
            warnings that it receives to Python's built-in `exceptions.warn()`
            function.
""")


selector_parameters = _strip("""
        selector (str or list[str] or callable or None): Indicates how to
            select the best candidate pipeline.

            If `selector` is a string, then it must be the name of a metric.
            In this case, autom8 will use that metric to select the best
            pipeline.

            If `selector` is a list of strings, then the strings must be the
            names of metrics. In this case, autom8 will use these metrics to
            select the best pipeline. (Metrics later in the list are used when
            earlier metrics are tied.)

            If `selector` is a callable object, then it should take two
            `autom8.Candidate` objects, and it should return the candidate that
            it thinks is better.

            If `selector` is None, then autom8 will use an appropriate default
            metric, based on the type of problem. For classification problems,
            autom8 will use the `f1_score` metrics. And for regression
            problems, autom8 will use `r2_score`.
""")


common_context_parameters = _render("""
        $dataset_parameters
        $problem_parameters
""")


all_context_parameters = _render("""
        $common_context_parameters
        $receiver_parameter
""")


def render_docstring(obj):
    obj.__doc__ = _render(obj.__doc__)
    return obj



from collections import namedtuple
import numpy as np
import scipy.sparse

from .matrix import create_matrix, Matrix
from .exceptions import Autom8Exception, expected, typename
from .preprocessors import playback
from .receiver import Receiver


PredictionReport = namedtuple('PredictionReport', 'predictions, probabilities')


class Pipeline:
    def __init__(self, input_columns, predicts_column, steps, estimator, label_encoder):
        self.input_columns = input_columns
        self.predicts_column = predicts_column
        self.steps = steps
        self.estimator = estimator
        self.label_encoder = label_encoder

    @property
    def predicts_classes(self):
        return self.label_encoder.classes_.tolist() if self.label_encoder else None

    def run(self, features, receiver=None):
        if not isinstance(features, (list, Matrix)):
            raise expected('list or Matrix', typename(features))

        if receiver is None:
            receiver = Receiver()

        if isinstance(features, list):
            features = create_matrix(features, receiver=receiver)

        # Rearrange the matrix so that the columns are in the right order.
        assert isinstance(features, Matrix)
        selected = self._select_columns(features, receiver)

        ctx = PlaybackContext(selected, receiver)
        playback(self.steps, ctx)
        return self._predict(ctx.matrix)

    def _predict(self, X):
        # TODO: Require a receiver, and notify it when the features need a lot
        # of coercion.
        X = X._float_array()

        has_proba = hasattr(self.estimator, 'predict_proba')
        probabilities = [] if has_proba else None
        predictions = []

        # TODO: Calculate an appropriate stride.
        stride = 1000
        for i in range(0, len(X), stride):
            window = scipy.sparse.csr_matrix(X[i : i + stride])

            if has_proba:
                y = self.estimator.predict_proba(window)
                rows = [self._format_probabilities(row) for row in y]
                predictions.extend(row[0][0] for row in rows)
                probabilities.extend(rows)
            else:
                y = self.estimator.predict(window)
                if self.label_encoder is not None:
                    y = self.label_encoder.inverse_transform(y)
                predictions.extend(y)

        return PredictionReport(
            _literalize(predictions),
            _literalize(probabilities),
        )

    def _format_probabilities(self, probs):
        classes = self.label_encoder.classes_
        pairs = [(c, p) for c, p in zip(classes, probs) if p > 0]
        return sorted(pairs, reverse=True, key=lambda pair: pair[1])[:3]

    def _select_columns(self, matrix, receiver):
        assert isinstance(matrix, Matrix)

        c1 = len(matrix.columns)
        c2 = len(self.input_columns)

        if c1 < c2:
            raise expected(f'matrix with at least {c2} columns', c1)

        try:
            return matrix.select_columns_by_name(self.input_columns)
        except Autom8Exception:
            pass

        receiver.warn(
            'Cannot match column names'
            f' from {[col.name for col in matrix.columns]}'
            f' to {self.input_columns}.'
        )

        if c1 == c2:
            return matrix
        else:
            return Matrix([col.copy() for col in matrix.columns[:c2]])


class PlaybackContext:
    def __init__(self, matrix, receiver):
        self.matrix = matrix
        self.receiver = receiver
        self.is_recording = False


def _literalize(obj):
    if hasattr(obj, 'tolist'):
        return obj.tolist()

    if isinstance(obj, list):
        return [_literalize(i) for i in obj]

    if isinstance(obj, tuple):
        return tuple(_literalize(i) for i in obj)

    return obj



import numpy as np
import sklearn.metrics

from .exceptions import expected
from .compare import better_candidate


class Candidate:
    def __init__(self, pipeline, formulas, train, test):
        self.pipeline = pipeline
        self.formulas = formulas
        self.train = train
        self.test = test

    def is_better_than(self, other):
        """
        Indicates if autom8 thinks this candidate is better than the other one.
        """
        return better_candidate(self, other) is self


class MetricsReport:
    def __init__(self, predictions, probabilities, metrics):
        self.predictions = predictions
        self.probabilities = probabilities
        self.metrics = metrics


def create_candidate(ctx, pipeline):
    """Creates a Candidate object from the provided context and pipeline.

    Parameters:
        ctx (RecordingContext): The current context.
        pipeline (Pipeline): The newly created pipeline.

    Returns:
        Candidate: The candidate object, containing the pipeline and its metrics.
    """

    return Candidate(
        pipeline=pipeline,
        formulas=ctx.matrix.formulas,
        train=_evaluate_predictions(ctx, pipeline, *ctx.training_data()),
        test=_evaluate_predictions(ctx, pipeline, *ctx.testing_data()),
    )


def _evaluate_predictions(ctx, pipeline, X, y):
    outputs = pipeline._predict(X)

    if ctx.is_regression:
        metrics = _evaluate_regressor(ctx, y, outputs.predictions)
        metrics.update(_extra_regressor_metrics(ctx, pipeline, metrics, len(y)))
    else:
        encoder = pipeline.label_encoder
        # MAY: Eventually use the probabilities, too.
        predicted = encoder.transform(outputs.predictions)
        metrics = _evaluate_classifier(ctx, y, predicted, encoder)

    return MetricsReport(outputs.predictions, outputs.probabilities, metrics)


def _evaluate_regressor(ctx, actual, predicted):
    funcs = [
        sklearn.metrics.explained_variance_score,
        sklearn.metrics.mean_absolute_error,
        sklearn.metrics.mean_squared_error,
        sklearn.metrics.mean_squared_log_error,
        sklearn.metrics.median_absolute_error,
        sklearn.metrics.r2_score,
    ]
    return _apply_metrics(funcs, actual, predicted)


def _extra_regressor_metrics(ctx, pipeline, initial_metrics, num_rows):
    funcs = [adjusted_r2_score]
    return _apply_metrics(funcs, ctx, pipeline, initial_metrics, num_rows)


def _evaluate_classifier(ctx, actual, predicted, encoder):
    funcs1 = [
        sklearn.metrics.f1_score,
        sklearn.metrics.precision_score,
        sklearn.metrics.recall_score,
    ]
    funcs2 = [
        sklearn.metrics.accuracy_score,
    ]
    funcs3 = [
        normalized_confusion_matrix,
        precision_recall_fscore_support,
    ]
    return {
        **_apply_metrics(funcs1, actual, predicted, average='weighted'),
        **_apply_metrics(funcs2, actual, predicted),
        **_apply_metrics(funcs3, actual, predicted, encoder),
    }


def _tolist(x):
    return x.tolist() if hasattr(x, 'tolist') else x


def _apply_metrics(funcs, *a, **k):
    result = {}
    for f in funcs:
        try:
            result[f.__name__] = _tolist(f(*a, **k))
        except Exception:
            pass
    return result


def normalized_confusion_matrix(actual, predicted, encoder):
    mat = sklearn.metrics.confusion_matrix(
        y_true=encoder.inverse_transform(actual),
        y_pred=encoder.inverse_transform(predicted),
        labels=encoder.classes_,
    )
    return mat.astype('float') / mat.sum(axis=1)[:, np.newaxis]


def precision_recall_fscore_support(actual, predicted, encoder):
    arrays = sklearn.metrics.precision_recall_fscore_support(
        y_true=encoder.inverse_transform(actual),
        y_pred=encoder.inverse_transform(predicted),
        labels=encoder.classes_,
    )
    result = []
    for cls, precision, recall, f1, support in zip(encoder.classes_, *arrays):
        result.append({
            'class': _tolist(cls),
            'precision': _tolist(precision),
            'recall': _tolist(recall),
            'f1': _tolist(f1),
            'support': _tolist(support),
        })
    return result


def adjusted_r2_score(ctx, pipeline, initial_metrics, num_rows):
    """Calculates the Adjusted R2 Score.

    See: https://en.wikipedia.org/wiki/Coefficient_of_determination#Adjusted_R2

    Parameters:
        ctx (RecordingContext): The current context.
        pipeline (Pipeline): The pipeline being evaluated.
        initial_metrics (dict): A dict of the metrics we've collected so far.
        num_rows (int): The number of rows in this segment of the dataset.

    Returns:
        float: The Adjusted R2 Score.
    """

    # Only count root columns that end up in a formula with a nonzero weight.
    # (A "root column" in a column that appeared in the initial matrix, before
    # any preprocessing, and was used to derive other columns.)
    roots = _nonzero_root_columns(ctx, pipeline)
    num_cols = len(roots)
    ratio = (num_rows - 1) / (num_rows - num_cols - 1)

    if ratio > 0:
        r2 = initial_metrics['r2_score']
        return 1 - (1 - r2) * ratio

    raise expected('more rows than columns',
        f'{num_rows} rows and {num_cols} columns.')


def _nonzero_root_columns(ctx, pipeline):
    est = pipeline.estimator
    weights = getattr(est, 'coef_', getattr(est, 'feature_importances_', None))
    columns = ctx.matrix.columns
    nonzeros = columns if weights is None else [
        c for c, w in zip(columns, weights) if w != 0
    ]
    return {i for c in nonzeros for i in c.root_columns()}



from .compare import better_candidate
from .docstrings import render_docstring
from .receiver import Receiver


@render_docstring
def create_selector(selector=None):
    """"Returns an `autom8.Receiver` that selects the best candidate pipeline.

    Parameters:
        $selector_parameters
    """

    if selector is None:
        return Comparator(lambda a, b: better_candidate(a, b))

    if isinstance(selector, str):
        return create_selector([selector])

    if isinstance(selector, (list, tuple)):
        if all(isinstance(i, str) for i in selector):
            compare = lambda r1, r2: _optimize_metrics(selector, r1, r2)
            return create_selector(compare)

    if callable(selector):
        return Comparator(selector)

    raise TypeError(
        'selector must be None, a string, a list of strings, or a function.'
    )


class Comparator(Receiver):
    def __init__(self, compare):
        self._compare = compare
        self.best = None

    def receive_candidate(self, candidate):
        if self.best is None:
            self.best = candidate
        else:
            self.best = self._compare(self.best, candidate)


def _optimize_metrics(metrics, candidate1, candidate2):
    for metric in metrics:
        x1 = _read_metric(candidate1, metric)
        x2 = _read_metric(candidate2, metric)

        # If it's not a tie, then return the candidate with the better score.
        if x1 != x2:
            is_better = x1 > x2 if metric.endswith('_score') else x1 < x2
            return candidate1 if is_better else candidate2

    # The reigning champion wins ties.
    return candidate1


def _read_metric(candidate, metric):
    if metric == 'column_count':
        return len(candidate.formulas)
    elif metric == 'step_count':
        return len(candidate.pipeline.steps)
    else:
        return candidate.test.metrics[metric]



from collections import namedtuple
from contextlib import contextmanager
import logging
import random

import numpy as np
import scipy.sparse

from .candidate import create_candidate
from .categories import LabelEncoder
from .cleaning import clean_numeric_labels
from .docstrings import render_docstring
from .exceptions import expected, typename
from .inference import _infer_role
from .matrix import create_matrix, Matrix
from .pipeline import Pipeline
from .receiver import Receiver


@render_docstring
def create_context(
    dataset,
    column_names=None,
    column_roles=None,
    target_column=None,
    problem_type=None,
    test_ratio=None,
    random_state=None,
    allow_multicore=True,
    executor_class=None,
    receiver=None,
):
    """Returns a new `autom8.RecordingContext` object, ready to create pipelines.

    Note:
        If you are not writing your own preprocessing or training logic, then
        you do not need to create your own context. You can use either
        `autom8.fit()` or `autom8.run()`, either of which will automatically
        create a context for you.

    Parameters:
        $all_context_parameters
    """

    # Create a default receiver if the user didn't provide one.
    if receiver is None:
        receiver = Receiver()

    # Small optimization: don't copy the matrix if we don't need to.
    if isinstance(dataset, Matrix) and column_names is None and column_roles is None:
        matrix = dataset
    else:
        matrix = create_matrix(
            dataset=dataset,
            column_names=column_names,
            column_roles=column_roles,
            receiver=receiver,
        )

    num_cols = len(matrix.columns)

    if num_cols == 0:
        raise expected('dataset with more than one column', repr(dataset))

    if num_cols == 1:
        raise expected('dataset with more than one column', num_cols)

    if target_column is None or target_column == -1:
        target_column = num_cols - 1

    if not isinstance(target_column, (int, str)):
        raise expected(f'target_column to be an int or a str',
            typename(target_column))

    if isinstance(target_column, str):
        target_column = target_column.strip()
        for index, col in enumerate(matrix.columns):
            if target_column == col.name.strip():
                target_column = index
                break

    if isinstance(target_column, str):
        raise expected(f'target_column to be one of: {matrix.column_names}',
            repr(target_column))

    if isinstance(target_column, int) and target_column >= len(matrix.columns):
        raise expected(
            f'target_column to be valid column number (less than {num_cols})',
            target_column,
        )

    labelcol = matrix.columns[target_column]
    label_name, label_values = labelcol.name, labelcol.values
    matrix.drop_columns_by_index(target_column)

    if problem_type is None:
        role = _infer_role(labelcol, receiver)
        problem_type = 'regression' if role == 'numerical' else 'classification'

    valid_problems = {'regression', 'classification'}
    if not isinstance(problem_type, str) or problem_type not in valid_problems:
        raise expected(f'problem_type in {valid_problems}', repr(problem_type))

    if problem_type == 'regression':
        label_values = clean_numeric_labels(label_name, label_values, receiver)

    if problem_type == 'regression':
        labels = Labels(label_name, label_values, label_values, None)
    else:
        assert problem_type == 'classification'
        encoder = LabelEncoder()
        encoded = encoder.fit_transform(label_values)
        labels = Labels(label_name, label_values, encoded, encoder)

    if test_ratio is None:
        test_ratio = 0.2

    if not isinstance(test_ratio, float) or not (0 < test_ratio < 1):
        raise expected('test_ratio between 0.0 and 1.0', test_ratio)

    count = len(matrix)
    test_indices = sorted(random.sample(range(count), int(count * test_ratio) or 1))

    if executor_class is None:
        executor_class = SynchronousExecutor

    return RecordingContext(
        matrix=matrix,
        labels=labels,
        test_indices=test_indices,
        problem_type=problem_type,
        random_state=random_state,
        allow_multicore=allow_multicore,
        executor_class=executor_class,
        receiver=receiver,
    )


class RecordingContext:
    """Keep track of autom8's internal state as it generates candidate
    pipelines.

    Notes:
        You can get a reference to this object in your receiver's
        `receive_context` method. If you keep your reference to the context
        around, keep in mind that autom8 mutates the context as it generates
        pipelines. Specically, the `matrix` and `steps` attributes frequently
        change while `autom8.run()` is running.

    Attributes:
        matrix (Matrix): The current feature matrix.
        labels (Labels): The labels that we're trying to predict.

            This is essentially the target column, but the values may be
            encoded, depending on whether or the data is categorical or
            numerical.

        test_indices (list[int]): A list of indices. Indicates which rows in
            should be used in the test dataset.
        problem_type (str): Either `classification` or `regression`.
        allow_multicore (bool): Indicates if estimators may use multiple cores.
        executor_class (class): The executor class that autom8 should use when
            it wants to run tasks in parallel.
        receiver (Receiver): An object that receives out-of-band data, like
            candidate pipelines and warnings.
        steps (list[Step]): A list of all the preprocessing steps that have
            been applied to the feature matrix.
        pool (Executor): The current executor, for executing tasks in parallel.
        is_recording (bool, always True): Indicates that this is a
            `RecordingContext` as opposed to a `PlaybackContext`.
    """

    def __init__(
            self, matrix, labels, test_indices, problem_type,
            random_state, allow_multicore, executor_class, receiver,
        ):
        self.input_columns = matrix.column_names
        self.matrix = matrix.copy()
        self.labels = labels
        self.test_indices = test_indices
        self.problem_type = problem_type
        self.random_state = random_state
        self.allow_multicore = allow_multicore
        self.executor_class = executor_class
        self.receiver = receiver
        self.steps = []
        self.pool = None
        self.is_recording = True

    @property
    def is_classification(self):
        return self.problem_type == 'classification'

    @property
    def is_regression(self):
        return self.problem_type == 'regression'

    @property
    def predicts_column(self):
        return self.labels.name

    @property
    def predicts_classes(self):
        return self.labels.classes

    @property
    def random_state_kw(self):
        return ({} if self.random_state is None
            else {'random_state': self.random_state})

    def __lshift__(self, estimator):
        """Provides convenient syntax for calling submit_fit.

        So instead of:

            ctx.submit_fit(xgboost.XGBRegressor(n_jobs=-1, random_state=1))

        You can write:

            ctx << xgboost.XGBRegressor(n_jobs=-1, random_state=1)
        """

        self.submit_fit(estimator)

    def submit_fit(self, estimator):
        """Submits `self.fit(estimator)` to the current executor.

        If the context does not currently have an executor, then this method
        simply calls `self.fit(estimator)`.
        """

        if self.pool is None:
            self.fit(estimator)
        elif hasattr(self.pool, 'fit'):
            self.pool.fit(self, estimator)
        else:
            self.pool.submit(self.fit, estimator)

    def fit(self, estimator):
        X, y = self.training_data()
        X = scipy.sparse.csr_matrix(X._float_array())

        try:
            estimator.fit(X, y)
        except Exception:
            logging.getLogger('autom8').exception('Failure in fit method')
            return

        pipeline = Pipeline(
            input_columns=self.input_columns,
            predicts_column=self.predicts_column,
            steps=list(self.steps),
            estimator=estimator,
            label_encoder=self.labels.encoder,
        )

        candidate = create_candidate(self, pipeline)
        self.receiver.receive_candidate(candidate)

    def submit(self, func, *args, **kwargs):
        if self.pool is None:
            func(*args, **kwargs)
        else:
            self.pool.submit(f, *args, **kwargs)

    def testing_data(self):
        feat = self.matrix.select_rows(self.test_indices)
        lab = self.labels.encoded[self.test_indices]
        return (feat, lab)

    def training_data(self):
        feat = self.matrix.exclude_rows(self.test_indices)
        lab = np.delete(self.labels.encoded, self.test_indices)
        return (feat, lab)

    @contextmanager
    def sandbox(self):
        saved_matrix = self.matrix.copy()
        saved_steps = list(self.steps)
        yield
        self.matrix = saved_matrix
        self.steps = saved_steps

    @contextmanager
    def parallel(self):
        if self.pool is not None:
            yield
            return

        num_steps = len(self.steps)
        self.pool = self.executor_class()
        yield

        try:
            self.pool.shutdown(wait=True)
        finally:
            self.pool = None

        if len(self.steps) != num_steps:
            self.receiver.warn(
                'Potential race condition: The RecordingContext was updated'
                ' within a `parallel` context. To avoid any race conditions,'
                ' use the `sandbox` method to create a temporary copy of the'
                ' RecordingContext. Then apply your preprocessing functions'
                ' to the sandboxed copy.'
            )


class Labels(namedtuple('Labels', 'name, original, encoded, encoder')):
    @property
    def classes(self):
        return self.encoder.classes_.tolist() if self.encoder else None



class SynchronousExecutor:
    def fit(self, context, estimator):
        self.submit(context.fit, estimator)

    def submit(self, func, *args, **kwargs):
        func(*args, **kwargs)

    def shutdown(self, wait=True):
        pass



import re
import numpy as np
from .exceptions import expected, typename
from .formats import parse_number
from .matrix import create_array
from .preprocessors import planner, preprocessor, _drop_weak_columns


int_type = (int, np.int64)


def clean_numeric_labels(name, values, receiver):
    if values.dtype in (float, int, bool):
        return values

    NumType = (float, int, np.int64, np.float64)
    if all(isinstance(i, NumType) for i in values):
        return values

    count = sum(1 for i in values if not isinstance(i, NumType))
    suffix = '' if count == 1 else 's'
    receiver.warn(
        f'Converting {count} value{suffix} in the "{name}" column'
        f' to number{suffix}.'
    )

    def conv(x):
        if isinstance(x, NumType):
            return x
        if isinstance(x, str):
            try:
                return parse_number(x)
            except Exception:
                pass
        return 0.0

    return np.array([conv(x) for x in values])


@planner
def clean_dataset(ctx):
    """Cleans the context's dataset, and records the steps for later playback.

    This function peforms a handful of steps:

    - Drops columns that contain unexpected values. A value is unexpected if is
      is not a boolean, a number, a string, or None.
    - Drops columns that only contain None values.
    - Whenever possible, coerces columns to have numeric types.
    - In columns of strings, replaces None values with the empty string.
    - In a column of numbers, if any numbers are missing, replaces the missing
      number with zero, and adds a second column of booleans to indicate which
      values were missing.
    - Replaces a column of mixed strings and numbers with two columns: one for
      the string values and another for the number values.

    After running these steps, each column in the dataset should have a uniform
    type for all the values in that column. For example, if a column has one
    boolean value, then all the values in the column will be booleans.

    Parameters:
        ctx (RecordingContext): The current context.
    """

    for col in ctx.matrix.columns:
        _clean_column(ctx, col)


def _clean_column(ctx, col):
    # If numpy inferred a real dtype for this column, then it's clean enough.
    # TODO: Consider bipartitioning columns with nan values.
    if col.dtype != object:
        return

    index = ctx.matrix.columns.index(col)
    values = col.values
    num_values = len(values)

    num_bools = sum(1 for i in values if isinstance(i, bool))
    num_floats = sum(1 for i in values if isinstance(i, float))

    num_ints = sum(1 for i in values
        if isinstance(i, int_type)
        and not isinstance(i, bool))

    num_none = sum(1 for i in values if i is None)
    num_strings = sum(1 for i in values if isinstance(i, str))
    computed_total = num_bools + num_floats + num_ints + num_none + num_strings

    if computed_total != num_values:
        try:
            found = {typename(i) for i in values if i is not None
                and not isinstance(i, (bool, int, float, str, np.int64))}
        except Exception:
            found = 'unexpected values'
        ctx.receiver.warn(f'Dropping column "{col.name}". A column must only'
            f' contain booleans, numbers, and strings. Received: {found}.')
        _drop_weak_columns(ctx, [index])
        return

    # Record the number of values for each type.
    counts = {bool: num_bools, float: num_floats, int: num_ints}

    # If we somehow got an array of primitives with dtype == object, then just
    # coerce it to the appropriate type.
    for typ, num in counts.items():
        if num == num_values:
            _coerce_column(ctx, index, typ)
            return

    # If we have all None values, then drop this column.
    if num_none == num_values:
        ctx.receiver.warn(f'Dropping column of all None values: {col.name}')
        _drop_weak_columns(ctx, [index])
        return

    # If we have some strings, see if we can convert them all to numbers.
    if num_strings > 0 and _can_coerce_all_strings_to_numbers(values):
        _coerce_strings_to_numbers(ctx, index)
        _clean_column(ctx, col)
        return

    # If we have all strings, then we're clean. Just leave this column alone.
    if num_strings == num_values:
        assert all(isinstance(i, str) for i in values)
        return

    # If we have all strings, but some are None, then use the empty string.
    # (And don't warn the user about the None values in this case.)
    if num_strings + num_none == num_values:
        _replace_none_values(ctx, index, '')
        return

    # If we have some ints and some floats, then coerce the ints to floats
    # and recur.
    if num_ints > 0 and num_floats > 0:
        _coerce_ints_to_floats(ctx, index)
        _clean_column(ctx, col)
        return

    # If any strings are blank, then replace them with the empty string and
    # recur.
    if any(i.isspace() for i in values if isinstance(i, str)):
        _replace_blank_strings(ctx, index)
        _clean_column(ctx, col)
        return

    # If all of the strings are the empty string, then replace the empty
    # strings with None and recur.
    if num_strings > 0 and all(i == '' for i in values if isinstance(i, str)):
        _replace_empty_strings(ctx, index, None)
        _clean_column(ctx, col)
        return

    # If we have all primitive values, but some are None, then replace None
    # values with the appropriate zero value. Add a boolean column that records
    # which values were missing.
    for typ, num in counts.items():
        if num + num_none == num_values:
            ctx.receiver.warn(
                f'Column {repr(col.name)} has {num_none} missing'
                f' value{"" if num_none == 1 else "s"}.'
            )
            _flag_missing_values(ctx, index, typ(0))
            return

    # Now we know we have some strings, some numbers, and maybe some Nones.
    # We also know that the nonempty strings cannot all be coerced into
    # numbers (but maybe some can be coerced).

    # First, coerce as many of the strings as we can into numbers. Then split
    # this column into two columns: one for numbers and another for strings.
    _coerce_strings_to_numbers(ctx, index)
    _bipartition_strings(ctx, index)


@preprocessor
def _coerce_column(ctx, index, to_type):
    """Converts all the items in a column to a specific data type."""
    col = ctx.matrix.columns[index]
    col.coerce(to_type)


_string_to_number_regex = re.compile(r'''
    ^\s* # leading spaces
    \$*  # any number of dollar sign characters
    (
        [\-\+]?  # optional sign
        (
            \.[0-9_]+      # decimal part without integral part
            |              # OR
            [0-9,_]+       # integral part
            (\.[0-9_]*)?   # and optional decimal part
        )
    )
    \%*  # any number of percent characters
    \s*$ # trailing spaces
''', re.VERBOSE)


def _can_coerce_all_strings_to_numbers(values):
    """Tests if each string in the collection can be coerced to a number.

    >>> _can_coerce_all_strings_to_numbers(['1', '2', '3'])
    True

    >>> _can_coerce_all_strings_to_numbers(['1.0', '$2', '3%', '.4', '5.', '-6', '+7'])
    True

    >>> _can_coerce_all_strings_to_numbers(['0,000', '1_234_567', '$-8.%', '_9'])
    True

    >>> _can_coerce_all_strings_to_numbers([None, 1, 2, '3'])
    True

    >>> _can_coerce_all_strings_to_numbers([None, 1, 2, '3', 'foo'])
    False

    >>> _can_coerce_all_strings_to_numbers([])
    True

    >>> _can_coerce_all_strings_to_numbers([None, (), object()])
    True

    >>> _can_coerce_all_strings_to_numbers(['1.2.3'])
    False

    >>> _can_coerce_all_strings_to_numbers(['1.2,'])
    False

    >>> _can_coerce_all_strings_to_numbers(['1+2'])
    False
    """
    return all(_string_to_number_regex.match(i)
        for i in values if isinstance(i, str))


@preprocessor
def _coerce_strings_to_numbers(ctx, index):
    """
    Converts textual data to numberical data whenever possible.

    For example, this function converts text like "$5" to the number 5.
    """
    col = ctx.matrix.columns[index]
    new_values = [_coerce_string_to_number(i)
        if isinstance(i, str) else i
        for i in col.values]
    col.values = create_array(new_values)


def _coerce_string_to_number(obj):
    assert isinstance(obj, str)

    # Remove leading and trailing whitespace.
    obj = obj.strip()

    # Replace the empty string with None.
    if obj == '':
        return None

    try:
        return parse_number(obj)
    except Exception:
        pass

    # Pull out the number part and try parsing it.
    try:
        m = _string_to_number_regex.match(obj)
        return parse_number(m.group(1)) if m else obj
    except Exception:
        # If we still can't parse it, then just admit defeat.
        return obj


@preprocessor
def _replace_none_values(ctx, index, replacement):
    """Treats missing data as blank text fields."""
    col = ctx.matrix.columns[index]
    new_values = [replacement if i is None else i for i in col.values]
    col.values = create_array(new_values)


@preprocessor
def _coerce_ints_to_floats(ctx, index):
    """Turns integer values into decimals, to make columns more uniform."""
    col = ctx.matrix.columns[index]
    new_values = [float(i) if isinstance(i, int_type) else i for i in col.values]
    col.values = create_array(new_values)


@preprocessor
def _replace_blank_strings(ctx, index):
    """Treats text that only contains space characters as blank fields."""
    col = ctx.matrix.columns[index]
    new_values = [
        '' if isinstance(i, str) and i.isspace() else i for i in col.values
    ]
    col.values = create_array(new_values)


@preprocessor
def _replace_empty_strings(ctx, index, replacement):
    """Treats blank text fields as missing data."""
    col = ctx.matrix.columns[index]
    new_values = [
        replacement if isinstance(i, str) and i == '' else i
        for i in col.values
    ]
    col.values = create_array(new_values)


@preprocessor
def _flag_missing_values(ctx, index, replacement):
    """
    Takes a column that contains missing data, and adds a new column
    of true/false values to indicate which rows contain missing data.
    """
    col = ctx.matrix.columns[index]
    new_values = [replacement if i is None else i for i in col.values]
    ctx.matrix.drop_columns_by_index([index])
    ctx.matrix.columns.append(col.copy_with(create_array(new_values)))
    ctx.matrix.append_column(
        values=col.values != None,
        formula=['is-defined', col],
        role='encoded',
        is_original=True,
    )


@preprocessor
def _bipartition_strings(ctx, index):
    """
    Takes a column that contains both text and numbers and turns it into two
    columns: one that contains only the text values, and another that contains
    only the numbers.
    """
    # For None values, just put a 0 in the number-column and an empty string
    # in the string-column.
    col = ctx.matrix.columns[index]
    ctx.matrix.drop_columns_by_index([index])

    is_num = lambda x: x is not None and not isinstance(x, str)
    numbers = [i if is_num(i) else 0 for i in col.values]
    strings = [i if isinstance(i, str) else '' for i in col.values]

    ctx.matrix.append_column(
        values=create_array(numbers),
        formula=['number', col],
        role=None,
        is_original=True,
    )

    ctx.matrix.append_column(
        values=create_array(strings),
        formula=['string', col],
        role=None,
        is_original=True,
    )



import io
import json
import pickle
import pprint
from datetime import date
from string import Template
from zipfile import ZipFile

import numpy
import pandas
import sklearn
import scipy

try:
    import lightgbm
except ImportError:
    lightgbm = None

try:
    import xgboost
except ImportError:
    xgboost = None

from . import __version__
from .docstrings import render_docstring
from .matrix import create_matrix


@render_docstring
def create_example_input(pipeline, dataset, indices, receiver=None):
    """Creates an example input object that you can pass to `create_package`.

    Parameters:
        pipeline (Pipeline): An `autom8.Pipeline` object.
        dataset (list or Matrix): The dataset used to create the pipeline.
        indices (list[int]): A list of row indices. Indicates which rows to use
            in the example input.
        $receiver_parameter

    Returns:
        list: A list of rows that can be passed to ``autom8.create_package()``
        as the `example_input` argument.
    """

    matrix = create_matrix(dataset, receiver=receiver)
    target_column = matrix.column_names.index(pipeline.predicts_column)
    matrix.drop_columns_by_index(target_column)
    return [pipeline.input_columns] + matrix.select_rows(indices).tolist()


def create_package(package_name, pipeline, example_input, extra_notes=None):
    """Creates a Python package for an `autom8.Pipeline` object.

    This function creates a zip archive and returns it as a `bytes` object.
    The zip archive contains a Python project that runs a web service for
    making predictions with the provided pipeline.

    The zip archive contains a pickle file of the pipeline, along with a flask
    application, test script, Dockerfile, and Makefile.

    Parameters:
        package_name (str): The name to use for this package.

            This name will be used as the archive's top-level folder. It will
            also be used as the name of the Docker image and Docker container.

        pipeline (Pipeline): The pipeline. autom8 will serialize this object
            in the achive's `pipeline.pickle` file.

        example_input (list[row]): A list of rows to use as the example input.

            This list of rows will appear in the README file and in the
            unit tests.

        extra_notes (str or None): Optional content that will appear at the end
            of the generated README file. Defaults to None.

    Returns:
        bytes: The contents of the generated zip archive.

        You may write the bytes to disk as a zip file, or transmit the bytes
        over the network.
    """

    args = _template_args(package_name, pipeline, example_input, extra_notes)
    result = io.BytesIO()
    templates = {
        '.dockerignore': dockerignore,
        'Dockerfile': dockerfile,
        'LICENSE': license,
        'Makefile': makefile,
        'README.md': readme,
        'requirements.txt': requirements,
        'service.py': service,
        'tests.py': tests,
    }
    with ZipFile(result, 'w') as out:
        out.writestr(f'{package_name}/pipeline.pickle', pickle.dumps(pipeline))
        for file_name, template in templates.items():
            contents = Template(template).substitute(args).strip() + '\n'
            out.writestr(f'{package_name}/{file_name}', contents.encode('utf-8'))
    return result.getvalue()


def _template_args(package_name, pipeline, example_input, extra_notes):
    est = pipeline.estimator
    extra_packages = []

    if lightgbm is not None:
        if isinstance(est, (lightgbm.LGBMRegressor, lightgbm.LGBMClassifier)):
            extra_packages.append(f'lightgbm=={lightgbm.__version__}')

    if xgboost is not None:
        if isinstance(est, (xgboost.XGBRegressor, xgboost.XGBClassifier)):
            extra_packages.append(f'xgboost=={xgboost.__version__}')

    example_result = pipeline.run(example_input)
    example_output = {
        'columnName': f'Predicted {pipeline.predicts_column}',
        'predictions': example_result.predictions,
        'probabilities': example_result.probabilities,
    }

    if hasattr(example_input, 'tolist'):
        example_input = example_input.tolist()

    return {
        'DOCKER_NAME': package_name,
        'ESTIMATOR_CLASS': type(pipeline.estimator).__name__,
        'EXTRA_NOTES': extra_notes or '',
        'EXTRA_PACKAGES': '\n'.join(extra_packages),
        'PACKAGE_NAME': package_name,
        'PACKAGE_NAME_REPR': repr(package_name),
        'PREDICTED_COLUMN': pipeline.predicts_column,
        'PREDICTED_COLUMN_REPR': repr(f'Predicted {pipeline.predicts_column}'),
        'README_INPUT_COLUMNS': '\n'.join(f'  -  {c}'
            for c in pipeline.input_columns),
        'README_INPUT_EXAMPLE': json
            .dumps({'rows': example_input}, sort_keys=True, indent=2)
            .replace("'", '\'\"\'\"\'').replace('\n', '\n        '),
        'README_OUTPUT_EXAMPLE': json
            .dumps(example_output, sort_keys=True, indent=2)
            .replace('\n', '\n    '),
        'TEST_INPUT': pformat(example_input),
        'TEST_OUTPUT_PREDICTIONS': pformat(example_result.predictions),
    }


def pformat(obj):
    return pprint.pformat(obj).replace('\n', '\n  ')


dockerfile = '''
FROM python:3.6.4

WORKDIR /deploy
COPY requirements.txt /deploy
RUN pip3 install --no-cache-dir --disable-pip-version-check -r requirements.txt

COPY LICENSE /deploy
COPY service.py /deploy
COPY pipeline.pickle /deploy

CMD ["gunicorn", "service:app", "--workers=1", "--bind", "0.0.0.0:80"]
'''


dockerignore = '''
.pytest_cache
.venv
__pycache__
'''


license = '''
MIT License

Copyright (c) 2018 Machinate, Inc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


makefile = '''
# This file defines commands for running and testing your web service. Use the
# `make` program to run them (for example, type `make test` to run the unit
# tests).


# Removes any compiled Python files that may be lying around.
clean:
    rm -rf __pycache__/*.pyc


# Runs your web service in a Docker container. Maps local port 55155 to the
# container's port 80.
container: image
    docker run \\
        --name $DOCKER_NAME \\
        --rm \\
        -p 0.0.0.0:55155:80 \\
        $DOCKER_NAME


# Builds the Docker image for your web service.
image:
    docker build -t $DOCKER_NAME .


# Stops the Docker cotainer.
stop:
    docker stop $DOCKER_NAME


# Runs the unit tests. (But first it runs `make clean` and `make venv`.)
test: clean venv
    .venv/bin/pytest -s -v -W "ignore::PendingDeprecationWarning" tests.py


# Creates the `activate` script for your virtual environment.
venv: .venv/bin/activate


# Creates your virtual environment and installs the required dependencies.
.venv/bin/activate: requirements.txt
    test -d .venv || python3 -m venv .venv
    .venv/bin/pip install -U -r requirements.txt
    .venv/bin/pip install -U pytest
    touch .venv/bin/activate


# Tells `make` which commands don't actaully produce any files.
.PHONY: clean container image stop test
'''.replace('    ', '\t')


readme = '''
# $PACKAGE_NAME

This project provides a stateless web service that hosts a machine learning model.
The web service provides a `/predict` URL for running the model and predicting
`$PREDICTED_COLUMN` values.


## Files

This package includes a handful of files:

- `.dockerignore` -- Prevents Docker from loading unnecessary files into its context.
- `Dockerfile` -- Defines the Docker image for your web service.
- `LICENSE` -- The MIT license for this software.
- `Makefile` -- Defines commands for starting and stopping the container, and running tests.
- `pipeline.pickle` -- The Python pickle file of your machine learning model.
- `README.md` -- This file. Provides some basic documentation.
- `requirements.txt` -- The Python dependencies for the web service.
- `service.py` -- Defines a flask application.
- `tests.py` -- Defines unit tests for your application. Run them with `make test`.


## Model Summary

The machine learning model is serialized in the `pipeline.pickle` file.

- Estimator: $ESTIMATOR_CLASS
- Predicts: $PREDICTED_COLUMN
- Inputs:
$README_INPUT_COLUMNS


## Requirements

- make
- docker
- python3


## Commands

- `make container` -- Runs the web service in a docker container.
- `make image` -- Builds the docker image.
- `make stop` -- Stops the container.
- `make test` -- Runs the unit tests in a virtual environment.


## URLs

- `GET /` -- Simply returns the string `$PACKAGE_NAME`. May be used as a health-check.
- `GET /describe` -- Returns some metadata about the machine learning model.
- `POST /predict` -- Runs the model on the provided rows and returns the predictions.


## Using the Container

Run the command `make container` to start the web service. This command maps
local port 55155 to the container's port 80.

To get predictions, send a POST request to `/predict`:

    curl --header "Content-Type: application/json" \\
        --data '$README_INPUT_EXAMPLE' \\
        http://localhost:55155/predict

This will return:

    $README_OUTPUT_EXAMPLE


$EXTRA_NOTES
'''


requirements = f'''
autom8=={__version__}
flask==1.*
gunicorn==19.*
numpy=={numpy.__version__}
pandas=={pandas.__version__}
scikit-learn=={sklearn.__version__}
scipy=={scipy.__version__}
$EXTRA_PACKAGES
'''


# TODO: Include an autom8 receiver when calling pipeline.run.
service = '''
import pickle
import flask

app = flask.Flask(__name__)
app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True

with open('pipeline.pickle', 'rb') as f:
    pipeline = pickle.load(f)


@app.route('/', methods=['GET'])
def health_check():
    return $PACKAGE_NAME_REPR


@app.route('/describe', methods=['GET'])
def describe():
    return flask.jsonify({
        'inputColumns': pipeline.input_columns,
        'predictsColumn': pipeline.predicts_column,
        'predictsClasses': pipeline.predicts_classes,
        'estimatorClass': type(pipeline.estimator).__name__,
        'estimator': repr(pipeline.estimator),
    })


@app.route('/predict', methods=['POST'])
def predict():
    try:
        result = pipeline.run(flask.request.json['rows'])
        return flask.jsonify({
            'columnName': f'Predicted {pipeline.predicts_column}',
            'predictions': result.predictions,
            'probabilities': result.probabilities,
        })
    except Exception:
        app.logger.exception('Failed to make predictions')
        raise
'''


tests = '''
import pickle
import service


SAMPLE_INPUT = $TEST_INPUT

SAMPLE_PREDICTIONS = $TEST_OUTPUT_PREDICTIONS


def test_pipeline_run_method():
    with open('pipeline.pickle', 'rb') as f:
        pipeline = pickle.load(f)
    result = pipeline.run(SAMPLE_INPUT)
    assert result.predictions == SAMPLE_PREDICTIONS


def test_index():
    client = service.app.test_client()
    response = client.get('/')
    assert response.data.decode('utf-8') == $PACKAGE_NAME_REPR


def test_describe():
    client = service.app.test_client()
    response = client.get('/describe')
    doc = response.get_json()
    expected_keys = [
        'inputColumns',
        'predictsColumn',
        'predictsClasses',
        'estimatorClass',
        'estimator',
    ]
    for key in expected_keys:
        assert key in doc


def test_predict():
    client = service.app.test_client()
    response = client.post('/predict', json={'rows': SAMPLE_INPUT})
    doc = response.get_json()
    assert doc['columnName'] == $PREDICTED_COLUMN_REPR
    assert doc['predictions'] == SAMPLE_PREDICTIONS
'''



#!/usr/bin/env python
import re
from setuptools import setup

with open('autom8/__init__.py', encoding='utf-8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

with open('README.md', encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='autom8',
    version=version,
    url='https://github.com/jvs/autom8',
    description='Python AutoML library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT License',
    author='John K. Von Seggern',
    author_email='john@machinate.com',
    packages=['autom8'],
    zip_safe=True,
    platforms='any',
    python_requires='>=3.6.0',
    install_requires=[
        'chardet>=3.0.4',
        'numpy>=1.14.4',
        'pandas>=0.23.4',
        'scikit-learn>=0.19.1',
        'scipy>=1.1.0',
    ],
    tests_require=['coverage', 'pytest>=3'],
    extras_require={
        'ml': [
            'lightgbm>=2.1.2',
            'xgboost>=0.80',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
    keywords=['machine learning', 'autoML'],
)



import warnings


class Autom8Exception(Exception):
    """The class of all exceptions raised by autom8."""


class Autom8Warning(UserWarning):
    """The class of all warnings issued by autom8."""


def expected(expected, received):
    return Autom8Exception(f'Expected {expected}. Received: {received}')


def typename(obj):
    try:
        return type(obj).__name__
    except Exception:
        return 'unexpected object'


def warn(message):
    warnings.warn(Autom8Warning(message), stacklevel=4)



def better_candidate(a, b):
    """Returns the candidate that autom8 thinks is better."""

    # If they're the same object, then exit early.
    if a is b:
        return a

    # If one is None, then return the other one.
    if a is None or b is None:
        return a or b

    # MAY: Take that class of estimator into consideration. (Maybe autom8
    # should prefer some estimator classes over other ones.)

    get_score = lambda m: m.get('r2_score', m.get('f1_score', 0))
    score_a = get_score(a.test.metrics)
    score_b = get_score(b.test.metrics)

    # If one score is at least 5% better than the other, then it wins.
    if score_a >= (score_b * 1.05): return a
    if score_b >= (score_a * 1.05): return b

    # If we have the "mean_absolute_error" metric, then see if there's a big
    # difference between the two.
    error_a = a.test.metrics.get('mean_absolute_error', 0)
    error_b = b.test.metrics.get('mean_absolute_error', 0)

    # If one is at least 10% smaller than the other, then it wins.
    if error_a <= (error_b * 0.9): return a
    if error_b <= (error_a * 0.9): return b

    # All right, see if we have the "accuracy_score" metric. If one score is
    # at least 10% better than the other, then it wins.
    acc_a = a.test.metrics.get('accuracy_score', 0)
    acc_b = b.test.metrics.get('accuracy_score', 0)
    if acc_a >= (acc_b * 1.10): return a
    if acc_b >= (acc_a * 1.10): return b

    # Well now we're in a gray area. See if one is simpler than the other.
    feat_a = _count_features(a)
    feat_b = _count_features(b)

    # If one is much simpler than the other, then it wins.
    if feat_a <= (feat_b - 3): return a
    if feat_b <= (feat_a - 3): return b

    # Well we tried! Just let the better score win.
    return a if score_a > score_b else b


def _count_features(candidate):
    num_cols = len(candidate.formulas)

    est = candidate.pipeline.estimator
    weights = getattr(est, 'coef_', getattr(est, 'feature_importances_', None))

    # If we don't have any weights, then just set them all to one.
    if weights is None:
        weights = [1 for _ in candidate.formulas]

    # If we don't have enough weights for some reason, then our estimator might
    # be messed up. In this case, just assume all the columns matter.
    if num_cols != len(weights):
        return num_cols

    for formula, weight in zip(candidate.formulas, weights):
        # Don't count any column whose importance is very close to zero.
        if abs(w) <= 0.0001:
            num_cols -= 1

        # Only count "encode-text" columns as one column.
        if isinstance(f, list) and f[0] == 'encode-text':
            num_cols -= 1
            saw_text = True

    # If we saw text columns, add one back to represent them all.
    if saw_text:
        num_cols += 1

    return num_cols
